package eu.olympus.client;

import eu.olympus.client.interfaces.UserClient;
import eu.olympus.model.Attribute;
import eu.olympus.model.PasswordAuthentication;
import eu.olympus.model.Policy;
import eu.olympus.model.exceptions.AuthenticationFailedException;
import eu.olympus.model.exceptions.UserCreationFailedException;
import eu.olympus.model.server.rest.IdentityProof;
import eu.olympus.server.PasswordJWTIdP;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import org.apache.commons.codec.binary.Base64;

public class PasswordJWTClient implements UserClient {

	private Map<Integer, PasswordJWTIdP> servers;
	protected Map<Integer, byte[]> sessionCookies;
	protected Map<Integer, Long> sessionStartedTimes;
	public static final long sessionLength = 6000000;
	public String password = null;
	
	public PasswordJWTClient(List<PasswordJWTIdP> servers) {
		this.servers = new HashMap<>();
		this.sessionCookies = new HashMap<>();
		this.sessionStartedTimes = new HashMap<>();
		int i = 0;
		for(PasswordJWTIdP server:servers){
			this.servers.put(i,server);
			sessionStartedTimes.put(i, 0l);
			i++;
		}
	}

	@Override
	public void createUser(String username, String password) throws UserCreationFailedException {
		servers.get(0).createUser(new PasswordAuthentication(username, password));
	}

	@Override
	public void createUserAndProveIdentity(String username, String password, IdentityProof identityProof) throws UserCreationFailedException {
		servers.get(0).createUserAndProveId(new PasswordAuthentication(username, password), identityProof.getStringRepresentation());
	}

	@Override
	public void proveIdentity(String username, String password,
			IdentityProof identityProof, String token, String type) throws UserCreationFailedException {
		try {
			getMFASession(username, password, token, type);
			servers.get(0).proveId(new PasswordAuthentication(username, password), sessionCookies.get(0), identityProof.getStringRepresentation());
		} catch(Exception e) {
			throw new UserCreationFailedException(e);
		}
	}

	@Override
	public String authenticate(String username, String password, Policy policy, String token, String type) throws AuthenticationFailedException {
		getMFASession(username, password, token, type);
		String reply = servers.get(0).authenticate(new PasswordAuthentication(username, password), sessionCookies.get(0), policy);
		if(reply == null) {
			throw new AuthenticationFailedException("Failed to authenticate");
		}
		return reply;
	}

	@Override
	public Map<String, Attribute> getAllAttributes(String username, String password, String token, String type) throws AuthenticationFailedException {
		try {
			getMFASession(username, password, token, type);
			return servers.get(0).getAllAttributes(new PasswordAuthentication(username, password), sessionCookies.get(0));
		}catch(Exception e) {
			throw new AuthenticationFailedException(e);
		}
	}

	@Override
	public void deleteAttributes(String username, String password, List<String> attributes, String token, String type) throws AuthenticationFailedException{
		try {
			getMFASession(username, password, token, type);
			if(!servers.get(0).deleteAttribute(new PasswordAuthentication(username, password), sessionCookies.get(0), attributes)) {
				throw new AuthenticationFailedException("Authentication failed");
			};
		}catch(Exception e) {
			throw new AuthenticationFailedException(e);
		}
	}

	@Override
	public void deleteAccount(String username, String password, String token, String type) throws AuthenticationFailedException {
		try{
			getMFASession(username, password, token, type);

			if (!servers.get(0).deleteAccount(new PasswordAuthentication(username, password), sessionCookies.get(0))) {
				throw new AuthenticationFailedException("Authentication failed");
			};
		}catch(Exception e) {
			throw new AuthenticationFailedException(e);
		}
	}

	@Override
	public void changePassword(String username, String oldPassword, String newPassword, String token, String type) throws UserCreationFailedException, AuthenticationFailedException{
		try {
			getMFASession(username, oldPassword, token, type);
			servers.get(0).changePassword(new PasswordAuthentication(username, oldPassword), new PasswordAuthentication(username, newPassword), sessionCookies.get(0));
		}catch(Exception e) {
			throw new AuthenticationFailedException(e);
		}
	}

	@Override
	public String requestMFAChallenge(String username, String password, String type)
			throws AuthenticationFailedException {
		return servers.get(0).requestMFA(new PasswordAuthentication(username, password), type);
	}

	@Override
	public void confirmMFA(String username, String password, String token, String type)
			throws AuthenticationFailedException {
		if(! servers.get(0).confirmMFA(new PasswordAuthentication(username, password), token, type) ) {
			throw new AuthenticationFailedException("Could not confirm MFA");
		}
	}

	@Override
	public void removeMFA(String username, String password, String token, String type)
			throws AuthenticationFailedException {
		if(! servers.get(0).removeMFA(new PasswordAuthentication(username, password), token, type) ) {
			throw new AuthenticationFailedException("Could not remove MFA");
		}
	}

	@Override
	public void clearSession() {
		this.sessionCookies.clear();
		this.sessionStartedTimes.clear();
		for(int i=0; i< servers.size(); i++){
			sessionStartedTimes.put(i, 0l);
		}
	}
	
	private void getMFASession(String username, String password, String token, String type) throws AuthenticationFailedException {
		if (System.currentTimeMillis() >= (sessionStartedTimes.get(0) + sessionLength)) {
			String sessionCookie = servers.get(0)
					.generateSession(new PasswordAuthentication(username, password), token, type);
			sessionCookies.put(0, Base64.decodeBase64(sessionCookie));
		}
		// Renew current session
		sessionStartedTimes.put(0, System.currentTimeMillis());
	}
}
