package eu.olympus.server;

import eu.olympus.model.Attribute;
import eu.olympus.model.PasswordAuthentication;
import eu.olympus.model.Policy;
import eu.olympus.model.exceptions.AuthenticationFailedException;
import eu.olympus.model.exceptions.UserCreationFailedException;
import eu.olympus.server.interfaces.VirtualIdP;
import eu.olympus.server.rest.Role;
import java.util.Arrays;
import java.util.List;
import java.util.Map;
import org.apache.commons.codec.binary.Base64;

public abstract class AbstractPasswordIdP implements VirtualIdP {

	protected PasswordHandler authenticationHandler;

	public void createUser(PasswordAuthentication creationData) throws UserCreationFailedException {
		authenticationHandler.createUser(creationData);
	}

	public void createUserAndProveId(PasswordAuthentication creationData, String idProof) throws UserCreationFailedException {
		createUser(creationData);
		addAttributes(creationData.getUsername(), idProof);	
	}

	public abstract String authenticate(PasswordAuthentication authenticationData, byte[] cookie, Policy policy) throws AuthenticationFailedException;

	public void proveId(PasswordAuthentication authenticationData, byte[] cookie, String idProof) throws UserCreationFailedException {
		if(validateSession(authenticationData, Base64.encodeBase64String(cookie))) {
			addAttributes(authenticationData.getUsername(), idProof);
		} else {
			throw new UserCreationFailedException();
		}
	}
	
	private void addAttributes(String username, String idProof) throws UserCreationFailedException {
		try {
			authenticationHandler.addAttributes(username, idProof);
		} catch(Exception e) {
			throw new UserCreationFailedException(e);
		}
	}

	public Map<String, Attribute> getAllAttributes(PasswordAuthentication authenticationData, byte[] cookie) throws AuthenticationFailedException {
		if(validateSession(authenticationData, Base64.encodeBase64String(cookie))) {
			return authenticationHandler.getAllAssertions(authenticationData.getUsername());
		} else {
			throw new AuthenticationFailedException("Authentication failed");
		}
	}

	public boolean deleteAttribute(PasswordAuthentication authenticationData, byte[] cookie, List<String> attributes) throws AuthenticationFailedException {
		if(validateSession(authenticationData, Base64.encodeBase64String(cookie))) {
			return authenticationHandler.deleteAttributes(authenticationData.getUsername(), attributes);
		}
		throw new AuthenticationFailedException("Authentication failed");
	}

	public boolean deleteAccount(PasswordAuthentication authenticationData, byte[] cookie) throws AuthenticationFailedException {
		if(validateSession(authenticationData, Base64.encodeBase64String(cookie))) {
			return authenticationHandler.deleteAccount(authenticationData.getUsername());
		}
		throw new AuthenticationFailedException("Authentication failed");
	}

	public void changePassword(PasswordAuthentication oldAuthenticationData, PasswordAuthentication newAuthenticationData, byte[] cookie) throws AuthenticationFailedException, UserCreationFailedException {
		if(validateSession(oldAuthenticationData, Base64.encodeBase64String(cookie))) {
			authenticationHandler.changePassword(oldAuthenticationData, newAuthenticationData);
		} else {
			throw new AuthenticationFailedException("Could not validate session");
		}
	}
	
	public String requestMFA(PasswordAuthentication authentication, String type) throws AuthenticationFailedException{
		try {
			boolean authenticated = authenticationHandler.authenticate(authentication);
			if(authenticated) {
				return authenticationHandler.requestMFASecret(authentication.getUsername(), type);
			}
		}catch (Exception e) {
		}
		throw new AuthenticationFailedException("Authentication failed");
	}
	
	public boolean confirmMFA(PasswordAuthentication authentication, String mfaToken, String type) {
		
		boolean authenticated = authenticationHandler.authenticate(authentication);
		if(authenticated) {
			if(authenticationHandler.validateMFAToken(authentication.getUsername(), mfaToken, type)) {
				return authenticationHandler.activateMFA(authentication.getUsername(), mfaToken, type);	
			}
		}
		return false;
	}

	public boolean removeMFA(PasswordAuthentication authentication, String mfaToken, String type) {
		boolean authenticated = authenticationHandler.authenticate(authentication);
		if(authenticated) {
			return authenticationHandler.deleteMFA(authentication.getUsername(), mfaToken, type);
		}
		return false;
	}

	public String generateSession(PasswordAuthentication authentication, String token, String type) throws AuthenticationFailedException{
		boolean tokenIsValid = authenticationHandler.validateMFAToken(authentication.getUsername(), token, type);
		if(tokenIsValid) {
			boolean authIsValid = authenticationHandler.authenticate(authentication);
			if (authIsValid) {
				return authenticationHandler.generateSessionCookie(authentication.getUsername());
			}
		}
		throw new AuthenticationFailedException("Failed to validate MFA token or password");
	}

	/**
	 * Validates that the password is correct and the cookie is valid
	 * @param authentication The credentials to validate
	 * @param cookie The cookie to validate
	 * @return True if the credentials and cookie are valid
	 */
	public boolean validateSession(PasswordAuthentication authentication, String cookie) {
		try {
			authenticationHandler.validateSession(cookie, Arrays.asList(Role.USER));
		} catch (AuthenticationFailedException e) {
			return false;
		}
		return authenticationHandler.authenticate(authentication);
	}
}
