package eu.olympus.server;

import eu.olympus.model.Authorization;
import eu.olympus.model.PasswordAuthentication;
import eu.olympus.model.exceptions.ExistingUserException;
import eu.olympus.model.exceptions.NonExistingUserException;
import eu.olympus.model.exceptions.UserCreationFailedException;
import eu.olympus.server.interfaces.MFAAuthenticator;
import eu.olympus.server.interfaces.ServerCryptoModule;
import eu.olympus.server.interfaces.Storage;
import eu.olympus.server.interfaces.UserPasswordDatabase;
import eu.olympus.server.rest.Role;
import eu.olympus.server.storage.InMemoryKeyDB;
import java.nio.charset.StandardCharsets;
import java.security.MessageDigest;
import java.util.Arrays;
import org.apache.commons.codec.binary.Base64;
import java.util.Map;

public class PasswordHandler extends AuthenticationHandler{

	private UserPasswordDatabase database;
	private MessageDigest md;
	private long sessionLength = 60000l;


	public PasswordHandler(Storage database,
			ServerCryptoModule cryptoModule, InMemoryKeyDB keyDB,
			Map<String, MFAAuthenticator> mfaAuthenticators) throws Exception {
		super(database, keyDB, mfaAuthenticators, cryptoModule);
		try {
			if(database instanceof UserPasswordDatabase) {
				this.database = (UserPasswordDatabase) database;
			} else {
				throw new Exception("Not a valid database");
			}
			md = MessageDigest.getInstance("SHA-512");
		} catch(Exception e) {
			throw new Exception("Setup failed", e);
		}
	}
	
	public void createUser(PasswordAuthentication creationData) throws UserCreationFailedException {
		if(this.database.hasUser(creationData.getUsername())) {
			throw new ExistingUserException();
		}
		byte[] saltBytes = crypto.getBytes(16);

		String hash = "";
		String salt = "";

		// Switch to a better suited hash algorithm
		md.reset();
		md.update(saltBytes);

		byte[] hashBytes = md.digest(creationData.getPassword().getBytes(StandardCharsets.UTF_8));

		salt = Base64.encodeBase64String(saltBytes);
		hash = Base64.encodeBase64String(hashBytes);
		this.database.addUser(creationData.getUsername(), salt, hash);
	}


	public void createUserAndProveId(PasswordAuthentication creationData, String idProof) throws UserCreationFailedException {
		createUser(creationData);
		addAttributes(creationData.getUsername(), idProof);
	}
	
	public boolean authenticate(PasswordAuthentication authenticationData) {
		String username =  authenticationData.getUsername();
		String password = this.database.getPassword(username);
		String salt = this.database.getSalt(username);
		String candidate = authenticationData.getPassword();
		if(password == null) {
			return false;
		}
		
		boolean valid = false;
	
		byte[] saltBytes = Base64.decodeBase64(salt);
			
		md.reset();
		md.update(saltBytes);
	
		byte[] hashBytes = md.digest(candidate.getBytes(StandardCharsets.UTF_8));
		String hash = Base64.encodeBase64String(hashBytes);
		valid = password.equals(hash);
		return valid;
	}
	
	public void changePassword(PasswordAuthentication old, PasswordAuthentication newData) throws UserCreationFailedException {
		if(authenticate(old)) {
			byte[] saltBytes = crypto.getBytes(16);
			
			String hash = "";
			String salt = "";

			// Switch to a better suited hash algorithm
			md.reset();
			md.update(saltBytes);

			byte[] hashBytes = md.digest(newData.getPassword().getBytes(StandardCharsets.UTF_8));

			salt = Base64.encodeBase64String(saltBytes);
			hash = Base64.encodeBase64String(hashBytes);
			this.database.setSalt(newData.getUsername(), salt);
			this.database.setPassword(newData.getUsername(), hash);
		} else {
			throw new UserCreationFailedException();
		}
	}

	@Override
	public String requestMFASecret(String username, String type) throws NonExistingUserException {
		if(!this.database.hasUser(username)) {
			throw new NonExistingUserException();
		}
		MFAAuthenticator authenticator = mfaAuthenticators.get(type);
		String secret = authenticator.generateSecret();
		this.database.assignMFASecret(username, type, secret);
		return secret;
	}


	@Override
	public String generateSessionCookie(String username) {
		String cookie = org.apache.commons.codec.binary.Base64.encodeBase64String(crypto.getBytes(64));
		storeAuthorization(cookie, new Authorization(username, Arrays.asList(Role.USER), System.currentTimeMillis()+this.sessionLength));
		return cookie;
	}

}
