package eu.olympus.server;

import eu.olympus.model.Attribute;
import eu.olympus.model.Authorization;
import eu.olympus.model.KeyShares;
import eu.olympus.model.OPRFResponse;
import eu.olympus.model.PabcPublicParameters;
import eu.olympus.model.Policy;
import eu.olympus.model.exceptions.AuthenticationFailedException;
import eu.olympus.model.exceptions.UserCreationFailedException;
import eu.olympus.server.interfaces.IdentityProver;
import eu.olympus.server.interfaces.MFAAuthenticator;
import eu.olympus.server.interfaces.PABCConfiguration;
import eu.olympus.server.interfaces.PestoIdP;
import eu.olympus.server.interfaces.ServerCryptoModule;
import eu.olympus.server.interfaces.Storage;
import eu.olympus.server.rest.Role;
import eu.olympus.server.storage.InMemoryKeyDB;
import eu.olympus.util.multisign.MSverfKey;
import java.security.PublicKey;
import java.security.cert.Certificate;
import java.util.Arrays;
import java.util.List;
import java.util.Map;
import org.apache.commons.codec.Charsets;
import org.apache.commons.codec.binary.Base64;
import org.miracl.core.BLS12461.ECP;

public class PestoIdPImpl implements PestoIdP {

	private int id;
	private ServerCryptoModule cryptoModule;
	private PestoAuthenticationHandler authenticationHandler;
	private ThresholdRSAJWTTokenGenerator tokenGenerator;
	private ThresholdPSSharesGenerator sharesGenerator;
	private Certificate cert;
	int nServers;
		
	public PestoIdPImpl(Storage database, List<IdentityProver> identityProvers, Map<String, MFAAuthenticator> authenticators, ServerCryptoModule cryptoModule) throws Exception {
		this.cryptoModule = cryptoModule;
		authenticationHandler = new PestoAuthenticationHandler(database, cryptoModule, new InMemoryKeyDB(), authenticators);
		tokenGenerator = new ThresholdRSAJWTTokenGenerator(cryptoModule);
		sharesGenerator = new ThresholdPSSharesGenerator(database, cryptoModule.getBytes(57));
		if (identityProvers != null) {
			for (IdentityProver idProver : identityProvers) {
				authenticationHandler.addIdentityProver(idProver);
			}
		}
	}

	public boolean setup(String ssid,
			PABCConfiguration pabcConfiguration,
			List<? extends PestoIdP> servers) {
		try {
			KeyShares master = new KeyShares(pabcConfiguration.getKeyMaterial(), pabcConfiguration.getRsaBlindings(),
					pabcConfiguration.getOprfKey(),pabcConfiguration.getOprfBlindings());
			boolean res = authenticationHandler.setup(ssid, master, pabcConfiguration.getLocalKeyShare(), pabcConfiguration.getRemoteShares(), pabcConfiguration.getId(),
					pabcConfiguration.getAllowedTimeDifference(), pabcConfiguration.getWaitTime(), pabcConfiguration.getSessionLength(), servers);
			id = pabcConfiguration.getId();
			cert = pabcConfiguration.getCert();
			sharesGenerator.setup(pabcConfiguration);
			nServers=servers.size()+1;

			return res;
		} catch(Exception e) {
			return false;
		}
	}
	
	@Override
	public void addSession(String cookie, Authorization authorization) {
		this.authenticationHandler.storeAuthorization(cookie, authorization);;
	}

	@Override
	public void validateSession(String cookie, List<Role> requestedRole) throws AuthenticationFailedException {
		this.authenticationHandler.validateSession(cookie, requestedRole);
	}

	@Override
	public OPRFResponse performOPRF(String ssid, String username, ECP x, String cookie) throws UserCreationFailedException, AuthenticationFailedException {
		validateSession(cookie, Arrays.asList(Role.USER));
		return authenticationHandler.performOPRF(ssid, username, x, cookie);
	}

	@Override
	public OPRFResponse performOPRF(String ssid, String username, ECP x, String mfaToken, String mfaType)
			throws UserCreationFailedException, AuthenticationFailedException {
		return authenticationHandler.performOPRF(ssid, username, x, mfaToken, mfaType);
	}
	
	public boolean startRefresh() {
		return authenticationHandler.startRefresh();
	}

	public void addMasterShare(String newSsid, byte[] share) {
		authenticationHandler.addMasterShare(newSsid, share);
	}

	public void setKeyShare(int id, byte[] newShare) {
		authenticationHandler.setKeyShare(id, newShare);
	}

	public void addPartialServerSignature(String ssid, byte[] signature) {
		authenticationHandler.addPartialServerSignature(ssid, signature);
	}

	public void addPartialMFASecret(String ssid, String secret, String type) {
		authenticationHandler.addPartialMFASecret(ssid, secret, type);
	}
	
	@Override
	public byte[] finishRegistration(String username, byte[] cookie, PublicKey publicKey, byte[] signature, long salt, String idProof) throws Exception {
		validateSession(Base64.encodeBase64String(cookie), Arrays.asList(Role.USER));
		return authenticationHandler.finishRegistration(username, cookie, publicKey, signature, salt, idProof);
	}
	
	@Override
	public String authenticate(String username, byte[] cookie, long salt, byte[] signature, Policy policy) throws Exception {
		validateSession(Base64.encodeBase64String(cookie), Arrays.asList(Role.USER));
		boolean authenticated = authenticationHandler.authenticate(username, cookie, salt, signature, AUTHENTICATESTRING.getBytes(Charsets.UTF_8));
		if(authenticated) {
			try{
				Map<String, Attribute> assertions = authenticationHandler
						.validateAssertions(username, policy);
				return tokenGenerator.generateToken(assertions);
			} catch(Exception e) {
				throw new AuthenticationFailedException("Failed : Could not produce a token");
			}
		}
		throw new AuthenticationFailedException("Failed : User failed authentication");
	}

	@Override
	public String obtainCredential(String username, byte[] cookie, long salt, byte[] signature, long timestamp) throws Exception {
		validateSession(Base64.encodeBase64String(cookie), Arrays.asList(Role.USER));
		boolean authenticated = authenticationHandler.authenticate(username, cookie, salt, signature, AUTHENTICATESTRING.getBytes(Charsets.UTF_8));
		if(authenticated) {
			return sharesGenerator.createCredentialShare(username,timestamp).getEncoded();
		}
		throw new AuthenticationFailedException("Failed : User failed authentication");
	}

	public MSverfKey getPabcPublicKey(){
		return sharesGenerator.getVerificationKeyShare();
	}

	public PabcPublicParameters getPabcPublicParam(){
		return sharesGenerator.getPublicParam();
	}

	@Override
	public Certificate getCertificate() {
		return cert;
	}

	@Override
	public int getId() {
		return id;
	}

	@Override
	public boolean proveIdentity(String username, byte[] cookie, long salt, byte[] signature, String idProof) throws AuthenticationFailedException {
		validateSession(Base64.encodeBase64String(cookie), Arrays.asList(Role.USER));
		boolean authenticated = authenticationHandler.authenticate(username, cookie, salt, signature, PROVEIDENTITYSTRING.getBytes(Charsets.UTF_8));
		try {
			if(authenticated) {
				this.authenticationHandler.addAttributes(username, idProof);
				return true;
			}
		} catch(Exception e) {
		}
		throw new AuthenticationFailedException("Failed : User failed authentication");
	}

	@Override
	public Map<String, Attribute> getAllAttributes(String username, byte[] cookie, long salt, byte[] signature) throws AuthenticationFailedException {
		validateSession(Base64.encodeBase64String(cookie), Arrays.asList(Role.USER));
		boolean authenticated = authenticationHandler.authenticate(username, cookie, salt, signature, GETALLATTRIBUTESSTRING.getBytes(Charsets.UTF_8));
		if(authenticated) {
			Map<String, Attribute> assertions = authenticationHandler
					.getAllAssertions(username);
			return assertions;
		}
		throw new AuthenticationFailedException("Failed : User failed authentication");
	}

	@Override
	public boolean deleteAttributes(String username, byte[] cookie, long salt, byte[] signature, List<String> attributes) throws AuthenticationFailedException {
		validateSession(Base64.encodeBase64String(cookie), Arrays.asList(Role.USER));
		boolean authenticated = authenticationHandler.authenticate(username, cookie, salt, signature, DELETEATTRIBUTESTRING.getBytes(Charsets.UTF_8));
		if(authenticated) {
			authenticationHandler.deleteAttributes(username, attributes);
			return true;
		}
		throw new AuthenticationFailedException("Failed : User failed authentication");
	}

	@Override
	public boolean deleteAccount(String username, byte[] cookie, long salt, byte[] signature) throws AuthenticationFailedException {
		validateSession(Base64.encodeBase64String(cookie), Arrays.asList(Role.USER));
		boolean authenticated = authenticationHandler.authenticate(username, cookie, salt, signature, DELETEACCOUNTSTRING.getBytes(Charsets.UTF_8));
		if(authenticated) {
			authenticationHandler.deleteAccount(username);
			return true;
		}
		throw new AuthenticationFailedException("Failed : User failed authentication");
	}

	@Override
	public byte[] changePassword(String username, byte[] cookie, PublicKey publicKey, byte[] oldSignature, byte[] newSignature, long salt) throws Exception {
		validateSession(Base64.encodeBase64String(cookie), Arrays.asList(Role.USER));
		return authenticationHandler.changePassword(username, cookie, publicKey, oldSignature, newSignature, salt);
	}

	@Override
	public String requestMFA(String username, byte[] cookie, long salt, String type, byte[] signature) throws AuthenticationFailedException{
		validateSession(Base64.encodeBase64String(cookie), Arrays.asList(Role.USER));
		try {
			boolean authenticated = authenticationHandler.authenticate(username, cookie, salt, signature, REQUESTMFASTRING.getBytes(Charsets.UTF_8));
			if(authenticated) {
				return authenticationHandler.requestMFASecret(username, type);
			}
		}catch (Exception e) {
		}
		throw new AuthenticationFailedException("Authentication failed");
	}
	
	@Override
	public boolean confirmMFA(String username, byte[] cookie, long salt, String token, String type, byte[] signature) throws AuthenticationFailedException {
		validateSession(Base64.encodeBase64String(cookie), Arrays.asList(Role.USER));
		boolean authenticated = authenticationHandler.authenticate(username, cookie, salt, signature, CONFIRMMFASTRING.getBytes(Charsets.UTF_8));
		if(authenticated) {
			return authenticationHandler.activateMFA(username, token, type);
		}
		throw new AuthenticationFailedException("Failed : User failed authentication");
	}

	@Override
	public boolean removeMFA(String username, byte[] cookie, long salt, String token, String type, byte[] signature) throws AuthenticationFailedException {
		validateSession(Base64.encodeBase64String(cookie), Arrays.asList(Role.USER));
		boolean authenticated = authenticationHandler.authenticate(username, cookie, salt, signature, REMOVEMFASTRING.getBytes(Charsets.UTF_8));
		if(authenticated) {
			return authenticationHandler.deleteMFA(username, token, type);
		}
		throw new AuthenticationFailedException("Failed : User failed authentication");
	}
	@Override
	public String refreshCookie(String cookie) {
		return authenticationHandler.refreshCookie(cookie);
	}
}
