package eu.olympus.server.rest;

import eu.olympus.model.PabcPublicParameters;
import eu.olympus.model.exceptions.AuthenticationFailedException;
import eu.olympus.server.interfaces.PestoIdP;
import eu.olympus.util.keyManagement.PemUtil;
import java.security.NoSuchAlgorithmException;
import java.security.PublicKey;
import javax.servlet.ServletContext;
import javax.ws.rs.Consumes;
import javax.ws.rs.GET;
import javax.ws.rs.HeaderParam;
import javax.ws.rs.POST;
import javax.ws.rs.Path;
import javax.ws.rs.Produces;
import javax.ws.rs.core.Context;
import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.Response;

import eu.olympus.model.server.rest.*;
import org.apache.commons.codec.binary.Base64;

import eu.olympus.model.OPRFResponse;
import eu.olympus.model.SerializedKey;
import eu.olympus.model.exceptions.UserCreationFailedException;
import eu.olympus.server.PestoIdPImpl;
import eu.olympus.util.KeySerializer;
import org.miracl.core.BLS12461.CONFIG_BIG;
import org.miracl.core.BLS12461.ECP;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.fasterxml.jackson.databind.ObjectMapper;

@Path("/idp")
public class PestoIdPServlet{

	private final String BEARERTOKENSTRING = "Bearer ";
	
	@Context ServletContext context;
	private static Logger logger = LoggerFactory.getLogger(PestoIdPServlet.class);

	@Path("requestOPRF")
	@POST
	@Produces(MediaType.APPLICATION_JSON)
	@Consumes(MediaType.APPLICATION_JSON)
	public OPRFRestResponse requestOPRF(OPRFRequest request) throws NoSuchAlgorithmException, UserCreationFailedException, AuthenticationFailedException {
		logger.info("idp/requestOPRF");
		logger.trace(getJson(request));
		PestoIdPImpl idp = (PestoIdPImpl) context.getAttribute("idp");
		ECP element = ECP.fromBytes(Base64.decodeBase64(request.getElement()));
		if(request.getSessionCookie()!= null) {
			OPRFResponse resp =	idp.performOPRF(request.getSsid(),
				request.getUsername(), element, request.getSessionCookie());
			byte[] fp12Bytes = new byte[12* CONFIG_BIG.MODBYTES];
			resp.getY().toBytes(fp12Bytes);
			return new OPRFRestResponse(resp.getSsid(), Base64.encodeBase64String(fp12Bytes), resp.getSessionCookie());
		} else {
			OPRFResponse resp = idp.performOPRF(request.getSsid(),
					request.getUsername(), element, request.getMfaToken(), request.getMfaType());
			byte[] fp12Bytes = new byte[12* CONFIG_BIG.MODBYTES];
			resp.getY().toBytes(fp12Bytes);
			return new OPRFRestResponse(resp.getSsid(), Base64.encodeBase64String(fp12Bytes), resp.getSessionCookie());
		}
		
	}
	
	@Path("requestMFA")
	@POST
	@Produces(MediaType.APPLICATION_JSON)
	@Consumes(MediaType.APPLICATION_JSON)
	public String requestMFA(SecondFactorRequest request) throws AuthenticationFailedException {
		logger.info("idp/requestMFA");
		logger.trace(getJson(request));
		PestoIdPImpl idp = (PestoIdPImpl) context.getAttribute("idp");
		return idp.requestMFA(request.getUsername(), Base64.decodeBase64(request.getSessionCookie()), request.getSaltIndex(), request.getType(), Base64.decodeBase64(request.getSignature()));
	}
	
	@Path("confirmMFA")
	@POST
	@Consumes(MediaType.APPLICATION_JSON)
	public Boolean confirmMFA(SecondFactorConfirmation request) throws AuthenticationFailedException {
		logger.info("idp/confirmMFA");
		logger.trace(getJson(request));
		PestoIdPImpl idp = (PestoIdPImpl) context.getAttribute("idp");
		return idp.confirmMFA(request.getUsername(), Base64.decodeBase64(request.getSessionCookie()), request.getSaltIndex(), request.getToken(), request.getType(), Base64.decodeBase64(request.getSignature()));
	}

	@Path("removeMFA")
	@POST
	@Consumes(MediaType.APPLICATION_JSON)
	public Boolean removeMFA(SecondFactorDelete request) throws AuthenticationFailedException {
		logger.info("idp/removeMFA");
		logger.trace(getJson(request));
		PestoIdPImpl idp = (PestoIdPImpl) context.getAttribute("idp");
		return idp.removeMFA(request.getUsername(), Base64.decodeBase64(request.getSessionCookie()), request.getSaltIndex(), request.getToken(), request.getType(), Base64.decodeBase64(request.getSignature()));
	}
	
	@Secured({Role.ADMIN})
	@Path("startRefresh")
	@POST
	@Consumes(MediaType.APPLICATION_JSON)
	public Boolean startRefresh() throws AuthenticationFailedException {
		logger.info("idp/startRefresh");
		PestoIdP idp = (PestoIdP) context.getAttribute("idp");
		return idp.startRefresh();
	}

	@Secured({Role.SERVER})
	@Path("addSignature")
	@POST
	@Consumes(MediaType.APPLICATION_JSON)
	public void addSignature(AddPartialSignatureRequest request) {
		logger.info("idp/addSignature");
		logger.trace(getJson(request));
		PestoIdP idp = (PestoIdP) context.getAttribute("idp");
		idp.addPartialServerSignature(request.getSsid(), Base64.decodeBase64(request.getString()));
	}

	@Secured({Role.SERVER})
	@Path("addSecret")
	@POST
	@Consumes(MediaType.APPLICATION_JSON)
	public void addMFASecret(AddPartialMFARequest request) {
		logger.info("idp/addSecret");
		logger.trace(getJson(request));
		PestoIdP idp = (PestoIdP) context.getAttribute("idp");
		idp.addPartialMFASecret(request.getSsid(), request.getString(), request.getType());
	}
	
	@Secured({Role.SERVER})
	@Path("setShares")
	@POST
	@Consumes(MediaType.APPLICATION_JSON)
	public void setShares(SetKeyShare request) {
		logger.info("idp/setShares");
		logger.trace(getJson(request));
		PestoIdP idp = (PestoIdP) context.getAttribute("idp");
		idp.setKeyShare(request.getId(), Base64.decodeBase64(request.getShares()));
	}

	@Secured({Role.SERVER})
	@Path("addMasterShare")
	@POST
	@Consumes(MediaType.APPLICATION_JSON)
	public Response addMasterShare(AddMasterShare request, @HeaderParam(value = "Authorization") String authorization) {
		logger.info("idp/addMasterShare");
		logger.trace(getJson(request));
		PestoIdP idp = (PestoIdP) context.getAttribute("idp");
		String newCookie = idp.refreshCookie(authorization.substring(BEARERTOKENSTRING.length()));
		try {
			idp.addMasterShare(request.getNewSsid(), Base64.decodeBase64(request.getNewShare()));
		} catch(Exception e) {
			logger.info("addMasterShare failed.", e);
			return Response.serverError().header("Authorization", "Authorization "+newCookie).build();
		}
		return Response.noContent().header("Authorization", BEARERTOKENSTRING+newCookie).build();
	}
	
	@Path("finishRegistration")
	@POST
	@Produces(MediaType.APPLICATION_JSON)
	@Consumes(MediaType.APPLICATION_JSON)
	public String finishRegistration(FinishRegistrationRequest request) throws Exception {
		logger.info("idp/finishRegistration");
		logger.trace(getJson(request));
		PestoIdP idp = (PestoIdP) context.getAttribute("idp");
		return Base64.encodeBase64String(idp.finishRegistration(request.getUsername(),
				Base64.decodeBase64(request.getSessionCookie()), (PublicKey)KeySerializer.deSerialize(request.getPublicKey()), Base64.decodeBase64(request.getSignature()),
				request.getSalt(), request.getIdProof()));
	}

	@Path("authenticate")
	@POST
	@Produces(MediaType.APPLICATION_JSON)
	@Consumes(MediaType.APPLICATION_JSON)
	public String authenticate(SignatureAndPolicy request) throws Exception {
		logger.info("idp/authenticate");
		logger.trace(getJson(request));
		PestoIdP idp = (PestoIdP) context.getAttribute("idp");
		return idp.authenticate(request.getUsername(), Base64.decodeBase64(request.getSessionCookie()),
				request.getSaltIndex(), Base64.decodeBase64(request.getSignature()),
				request.getPolicy());
	}

	@Path("getcertificate")
	@GET
	@Produces(MediaType.APPLICATION_JSON)
	public String getCertificate() throws Exception {
		logger.info("idp/getcertificate");
		PestoIdP idp = (PestoIdP) context.getAttribute("idp");
		return PemUtil.encodeDerToPem(idp.getCertificate().getEncoded(), "CERTIFICATE");
	}

	@Path("proveId")
	@POST
	@Consumes(MediaType.APPLICATION_JSON)
	public Boolean proveId(SignatureAndIdProof request) throws Exception {
		logger.info("idp/proveId");
		logger.trace(getJson(request));
		PestoIdP idp = (PestoIdP) context.getAttribute("idp");
		return idp.proveIdentity(request.getUsername(), Base64.decodeBase64(request.getSessionCookie()), request.getSaltIndex(),
			Base64.decodeBase64(request.getSignature()), request.getIdProof());
	}

	@Path("getAllAttributes")
	@POST
	@Consumes(MediaType.APPLICATION_JSON)
	@Produces(MediaType.APPLICATION_JSON)
	public AttributeMap getAllAttributes(GetAllAttributesRequest request) throws Exception {
		logger.info("idp/getAllAttributes");
		logger.trace(getJson(request));
		PestoIdP idp = (PestoIdP) context.getAttribute("idp");
		return new AttributeMap(idp.getAllAttributes(request.getUsername(), Base64.decodeBase64(request.getSessionCookie()), request.getSaltIndex(),
			Base64.decodeBase64(request.getSignature())));
	}
	
	@Path("deleteAttributes")
	@POST
	@Consumes(MediaType.APPLICATION_JSON)
	public Boolean deleteAttributes(DeleteAttributesRequest request) throws Exception {
		logger.info("idp/deleteAttributes");
		logger.trace(getJson(request));
		PestoIdP idp = (PestoIdP) context.getAttribute("idp");
		return idp.deleteAttributes(request.getUsername(), Base64.decodeBase64(request.getSessionCookie()), request.getSaltIndex(),
			Base64.decodeBase64(request.getSignature()), request.getAttributes());
	}
	
	@Path("changePassword")
	@POST
	@Produces(MediaType.APPLICATION_JSON)
	@Consumes(MediaType.APPLICATION_JSON)
	public String changePassword(ChangePasswordRequest request) throws Exception {
		logger.info("idp/changePassword");
		logger.trace(getJson(request));
		PestoIdP idp = (PestoIdP) context.getAttribute("idp");
		return Base64.encodeBase64String(idp.changePassword(request.getUsername(),
				Base64.decodeBase64(request.getSessionCookie()), (PublicKey)KeySerializer.deSerialize(request.getPublicKey()), Base64.decodeBase64(request.getOldSignature()),
				Base64.decodeBase64(request.getNewSignature()), request.getSalt()));
	}
	
	@Path("deleteAccount")
	@POST
	@Consumes(MediaType.APPLICATION_JSON)
	public Boolean deleteAccount(DeleteAccountRequest request) throws Exception {
		logger.info("idp/deleteAccount");
		logger.trace(getJson(request));
		PestoIdP idp = (PestoIdP) context.getAttribute("idp");
		return idp.deleteAccount(request.getUsername(), Base64.decodeBase64(request.getSessionCookie()), request.getSaltIndex(),
			Base64.decodeBase64(request.getSignature()));
	}
	
	@Path("getCredentialShare")
	@POST
	@Produces(MediaType.APPLICATION_JSON)
	@Consumes(MediaType.APPLICATION_JSON)
	public String getPSCredentialShare(SignatureAndTimestamp request) throws Exception {
		logger.info("idp/getCredentialShare");
		logger.trace(getJson(request));
		PestoIdPImpl idp = (PestoIdPImpl) context.getAttribute("idp");
		return idp.obtainCredential(request.getUsername(),
				Base64.decodeBase64(request.getSessionCookie()), request.getSaltIndex(), Base64.decodeBase64(request.getSignature()),request.getTimestamp());
	}

	@Path("getPabcPublicKeyShare")
	@GET
	@Produces(MediaType.APPLICATION_JSON)
	public SerializedKey getPABCPublicKeyShare() {
		logger.info("idp/getPabcPublicKeyShare");
		PestoIdPImpl idp = (PestoIdPImpl) context.getAttribute("idp");
		return KeySerializer.serialize(idp.getPabcPublicKey());
	}

	@Path("getPabcPublicParam")
	@GET
	@Produces(MediaType.APPLICATION_JSON)
	public PabcPublicParameters getPABCPublicParam() {
		logger.info("idp/getPabcPublicParam");
		PestoIdPImpl idp = (PestoIdPImpl) context.getAttribute("idp");
		return idp.getPabcPublicParam();
	}
	
	protected String getJson(Object obj) {
		try {
			return new ObjectMapper().writeValueAsString(obj);
		} catch(Exception e) {
			return "Could not convert "+obj+": "+e.getLocalizedMessage();
		}
	}
}