package eu.olympus.client;

import eu.olympus.model.Attribute;
import eu.olympus.model.PasswordAuthentication;
import eu.olympus.model.Policy;
import eu.olympus.model.SerializedKey;
import eu.olympus.model.exceptions.AuthenticationFailedException;
import eu.olympus.model.exceptions.UserCreationFailedException;
import eu.olympus.model.server.rest.AttributeMap;
import eu.olympus.model.server.rest.KeyAndCert;
import eu.olympus.model.server.rest.PasswordAuthenticationAndAttributes;
import eu.olympus.model.server.rest.PasswordAuthenticationAndCookie;
import eu.olympus.model.server.rest.PasswordAuthenticationAndIDProof;
import eu.olympus.model.server.rest.PasswordAuthenticationAndMFAToken;
import eu.olympus.model.server.rest.PasswordAuthenticationAndMFAType;
import eu.olympus.model.server.rest.PasswordAuthenticationAndPassword;
import eu.olympus.model.server.rest.PasswordAuthenticationAndPolicy;
import eu.olympus.server.PasswordJWTIdP;
import eu.olympus.server.interfaces.IdentityProver;
import eu.olympus.util.KeySerializer;
import eu.olympus.util.keyManagement.CertificateUtil;
import eu.olympus.util.keyManagement.PemUtil;
import java.security.cert.Certificate;
import java.security.interfaces.RSAPrivateKey;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import javax.ws.rs.client.Client;
import javax.ws.rs.client.ClientBuilder;
import javax.ws.rs.client.Entity;
import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.Response;
import org.apache.commons.codec.binary.Base64;
import org.eclipse.jetty.http.HttpStatus;


/**
 * REST wrapper for the IdP
 * @param <T>
 *
 */
public class PasswordIdPRESTConnection extends PasswordJWTIdP {

	private String host;
	private Client client;
	
	/**
	 * Create a new rest connections to an IdP
	 * @param url includes port, eg. http://127.0.0.1:9090
	 */
	public PasswordIdPRESTConnection(String url) {
		super(null, new LinkedList<IdentityProver>(), null);
		this.host = url+"/idp/";
	    this.client = ClientBuilder.newClient();
	}
	
	@Override
	public void setup(RSAPrivateKey privKey, Certificate certificate) throws Exception {
		SerializedKey serializedKey = KeySerializer.serialize(privKey);
		String pemEncodedCert = PemUtil.encodeDerToPem(certificate.getEncoded(), "CERTIFICATE");
		KeyAndCert pair = new KeyAndCert(serializedKey, pemEncodedCert);
	    client.target(host+"setup").request().post(Entity.entity(pair, MediaType.APPLICATION_JSON));
	}

	@Override
	public Certificate getCertificate() {
		try {
			String output = client.target(host + "getcertificate").request().get(String.class);
			return CertificateUtil.decodePemCert(output);
		} catch (Exception e) {
			throw new RuntimeException(e);
		}
	}

	@Override
	public String authenticate(PasswordAuthentication authenticationData, byte[] cookie, Policy policy) throws AuthenticationFailedException {
		try {
			PasswordAuthenticationAndPolicy data = new PasswordAuthenticationAndPolicy((PasswordAuthentication)authenticationData,
					Base64.encodeBase64String(cookie), policy);
			return client.target(host+"authenticate").request().post(Entity.entity(data, MediaType.APPLICATION_JSON), String.class);
		} catch (Exception e) {
	    	throw new AuthenticationFailedException(e);
	    }
	}

	@Override
	public void createUser(PasswordAuthentication creationData) throws UserCreationFailedException {
		try{
			Response response = client.target(host+"createuser").request().post(Entity.entity(creationData, MediaType.APPLICATION_JSON));
		    if(response.getStatus() != 204) {
		    	throw new UserCreationFailedException();
		    }
		} catch (Exception e) {
			throw new UserCreationFailedException(e);
		}
	}

	@Override
	public void createUserAndProveId(PasswordAuthentication creationData, String idProof)
			throws UserCreationFailedException {
		try{
			PasswordAuthenticationAndIDProof passwordAuthenticationAndIDProof = new PasswordAuthenticationAndIDProof(creationData, null, idProof.toString());
			Response response = client.target(host+"createuserandprove").request().post(Entity.entity(passwordAuthenticationAndIDProof, MediaType.APPLICATION_JSON));
		    if(response.getStatus() != 204) {
		    	throw new UserCreationFailedException();
		    }
		} catch (Exception e) {
			throw new UserCreationFailedException(e);
		}
	}

	@Override
	public void proveId(PasswordAuthentication authenticationData, byte[] cookie, String idProof) throws UserCreationFailedException {
		try{
			PasswordAuthenticationAndIDProof pwAndProof = new PasswordAuthenticationAndIDProof((PasswordAuthentication)authenticationData, Base64.encodeBase64String(cookie), idProof.toString());
			Response response = client.target(host+"prove").request().post(Entity.entity(pwAndProof, MediaType.APPLICATION_JSON));

		    if(response.getStatus() != 204) {
		    	throw new UserCreationFailedException();
		    }
		} catch (Exception e) {
			throw new UserCreationFailedException(e);
		}
	}

	@Override
	public Map<String, Attribute> getAllAttributes(PasswordAuthentication authenticationData, byte[] cookie) throws AuthenticationFailedException {
	    try {
				PasswordAuthenticationAndCookie pwAndCookie = new PasswordAuthenticationAndCookie((PasswordAuthentication)authenticationData, Base64.encodeBase64String(cookie));
	    	AttributeMap attributes = client.target(host+"getAllAttributes").request().post(Entity.entity(pwAndCookie, MediaType.APPLICATION_JSON), AttributeMap.class);
	    	return attributes.getAttributes();
	    } catch (Exception e) {
	    	throw new AuthenticationFailedException(e);
	    }
	}

	@Override
	public boolean deleteAttribute(PasswordAuthentication authenticationData, byte[] cookie, List<String> attributes) throws AuthenticationFailedException {
		try {
			PasswordAuthenticationAndAttributes data = new PasswordAuthenticationAndAttributes(authenticationData, Base64.encodeBase64String(cookie), attributes);
			boolean response = client.target(host+"deleteAttribute").request().post(Entity.entity(data, MediaType.APPLICATION_JSON), Boolean.class);
			return response;
		} catch(Exception e) {
			throw new AuthenticationFailedException(e);
		}
	}

	@Override
	public boolean deleteAccount(PasswordAuthentication authenticationData, byte[] cookie) throws AuthenticationFailedException {
		try {
			PasswordAuthenticationAndCookie pwAndCookie = new PasswordAuthenticationAndCookie((PasswordAuthentication)authenticationData, Base64.encodeBase64String(cookie));
			return  client.target(host+"deleteAccount").request().post(Entity.entity(pwAndCookie, MediaType.APPLICATION_JSON), Boolean.class);
		} catch(Exception e) {
			throw new AuthenticationFailedException(e);
		}
	}

	@Override
	public void changePassword(PasswordAuthentication oldAuthenticationData, PasswordAuthentication newAuthenticationData, byte[] cookie) throws UserCreationFailedException, AuthenticationFailedException {
		PasswordAuthenticationAndPassword request = new PasswordAuthenticationAndPassword(oldAuthenticationData, Base64.encodeBase64String(cookie), newAuthenticationData.getPassword());
		Response resp;
		try{
			resp = client.target(host+"changePassword").request().post(Entity.entity(request, MediaType.APPLICATION_JSON));
		} catch (Exception e) {
			throw new UserCreationFailedException(e);
		}
		// If the response is not a Success
		if (!HttpStatus.getCode(resp.getStatus()).isSuccess()) {
			throw new AuthenticationFailedException("Could not change password");
		}
	}
	
	@Override
	public String requestMFA(PasswordAuthentication authentication, String type) throws AuthenticationFailedException {
		PasswordAuthenticationAndMFAType request = new PasswordAuthenticationAndMFAType(authentication, type);
		try {
			return client.target(host+"requestMFA").request().post(Entity.entity(request, MediaType.APPLICATION_JSON), String.class);
		}catch (Exception e) {
			throw new AuthenticationFailedException(e);
		}
	}
	
	@Override
	public boolean confirmMFA(PasswordAuthentication authentication, String mfaToken, String type) {
		PasswordAuthenticationAndMFAToken request = new PasswordAuthenticationAndMFAToken(authentication, mfaToken, type);
		
		try {
			return client.target(host+"confirmMFA").request().post(Entity.entity(request, MediaType.APPLICATION_JSON), Boolean.class);
		}catch (Exception e) {
			throw e;
		}
	}

	@Override
	public boolean removeMFA(PasswordAuthentication authentication, String mfaToken, String type) {
		PasswordAuthenticationAndMFAToken request = new PasswordAuthenticationAndMFAToken(authentication, mfaToken, type);

		try {
			return client.target(host+"removeMFA").request().post(Entity.entity(request, MediaType.APPLICATION_JSON), Boolean.class);
		}catch (Exception e) {
			throw e;
		}
	}

	@Override
	public String generateSession(PasswordAuthentication authentication, String token, String type)  throws AuthenticationFailedException {
		PasswordAuthenticationAndMFAToken request = new PasswordAuthenticationAndMFAToken(authentication, token, type);
		try {
			return client.target(host+"generateSession").request().post(Entity.entity(request, MediaType.APPLICATION_JSON), String.class);
		}catch (Exception e) {
			throw new AuthenticationFailedException(e);
		}
	}
}
