package eu.olympus.client;

import static org.hamcrest.core.Is.is;
import static org.hamcrest.core.IsCollectionContaining.hasItem;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertThat;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;

import com.fasterxml.jackson.databind.ObjectMapper;

import eu.olympus.TestParameters;
import eu.olympus.model.Attribute;
import eu.olympus.model.Operation;
import eu.olympus.model.PasswordAuthentication;
import eu.olympus.model.Policy;
import eu.olympus.model.Predicate;
import eu.olympus.model.exceptions.AuthenticationFailedException;
import eu.olympus.model.exceptions.UserCreationFailedException;
import eu.olympus.server.PasswordJWTIdP;
import eu.olympus.server.GoogleAuthenticator;
import eu.olympus.server.SoftwareServerCryptoModule;
import eu.olympus.server.interfaces.MFAAuthenticator;
import eu.olympus.server.interfaces.Storage;
import eu.olympus.server.storage.InMemoryUserPasswordDatabase;
import eu.olympus.unit.server.TestIdentityProof;
import java.io.IOException;
import java.security.cert.Certificate;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Random;
import org.junit.BeforeClass;
import org.junit.Test;

public class TestPasswordJWTClient {

	private final static String user = "username";
	private final static String password = "password";
	private final static String signature = "signature";
	private final static String challenge = "challenge";
	private final static Map<String, Attribute> attributes = new HashMap<>();
	private final static List<String> claims = new LinkedList<String>();
	
	static PasswordJWTIdP dummyIdP;

	static class TestPasswordJWTIdP extends PasswordJWTIdP {

		public TestPasswordJWTIdP(Storage db, HashMap<String, MFAAuthenticator> mfas ) {
			super(db, new ArrayList<>(), mfas);
		}

		ObjectMapper objectMapper = new ObjectMapper();

		@Override
		public Certificate getCertificate() {
			return null;
		}

		@Override
		public void createUser(PasswordAuthentication creationData)
					throws UserCreationFailedException {
			assertThat(creationData.getUsername(), is(user));
			assertThat(creationData.getPassword(), is(password));
		}

		@Override
		public void createUserAndProveId(PasswordAuthentication creationData, String proof)
					throws UserCreationFailedException {
			try {
				TestIdentityProof idProof = objectMapper.readValue(proof, TestIdentityProof.class);
				assertThat(creationData.getUsername(), is(user));
				assertThat(creationData.getPassword(), is(password));
				assertThat(idProof.getSignature(), is(signature));
				assertThat(idProof.getAttributes().size(), is(2));
				for (String s : idProof.getAttributes().keySet()) {
					assertThat(idProof.getAttributes().get(s), is(attributes.get(s)));
				}
			} catch (Exception e) {
				fail(e.getMessage());
			}
		}

		@Override
		public void proveId(PasswordAuthentication authenticationData, byte[] cookie, String proof)
					throws UserCreationFailedException {
			try {
				TestIdentityProof idProof = objectMapper.readValue(proof, TestIdentityProof.class);

				assertThat(authenticationData.getUsername(), is(user));
				assertThat(authenticationData.getPassword(), is(password));
				assertThat(idProof.getSignature(), is(signature));
				assertThat(idProof.getAttributes().size(), is(2));
				for (String s : idProof.getAttributes().keySet()) {
					assertThat(idProof.getAttributes().get(s), is(attributes.get(s)));
				}
			} catch (IOException e) {
				fail(e.getMessage());
			}
		}

		@Override
		public String authenticate(PasswordAuthentication authenticationData, byte[] cookie, Policy policy) {
			if (!authenticationData.getUsername().equals(user) || !authenticationData.getPassword().equals(password)) {
				return null;
			}
			for (Predicate p : policy.getPredicates()) {
				assertThat(claims, hasItem(p.getAttributeName()));
			}
			assertEquals(policy.getPredicates().size(), claims.size());
			return "Dummy-Authenticate";
		}

		@Override
		public String generateSession(PasswordAuthentication auth, String token, String type) {
			if (!auth.getUsername().equals(user) || !auth.getPassword().equals(password)) {
				return "invalid-cookie";
			}
			return "valid-cookie";
		}

		@Override
		public boolean deleteAccount(PasswordAuthentication authenticationData, byte[] cookie) throws AuthenticationFailedException {
			if (!authenticationData.getUsername().equals(user) || !authenticationData.getPassword().equals(password)) {
				return false;
			}
			return true;
		}

		@Override
		public boolean deleteAttribute(PasswordAuthentication authenticationData, byte[] cookie, List<String> attributes) throws AuthenticationFailedException {
			if (!authenticationData.getUsername().equals(user) || !authenticationData.getPassword().equals(password)) {
				return false;
			}
			return true;
		}

		@Override
		public String requestMFA(PasswordAuthentication authentication, String type) throws AuthenticationFailedException{
			assertThat(authentication.getUsername(), is(user));
			assertThat(authentication.getPassword(), is(password));
			return challenge;
		}
	};

	static class TestInMemoryStorage extends InMemoryUserPasswordDatabase {
		@Override
		public boolean hasUser(String username) {
			if (username.equals(user)) {
				return true;
			}
			return false;
		}

		@Override
		public String getSalt(String username) {
			return "salt";
		}

		@Override
		public String getPassword(String username) {
			if (username.equals(username)) {
				return "h50rZTkC3PWiu4hjyvowdht3BzuR4EkTW4dbfYSABVbJ7vLVZwlO8Mihj4HqLTOxoo15aWkOPRVCpASig0qRcw=="; // hashed password
			}
			return null;
		}
	};

	@BeforeClass
	public static void setup(){
		claims.add("name");
		claims.add("age");
		attributes.put("name", new Attribute("John"));
		attributes.put("age", new Attribute(22));
		HashMap<String, MFAAuthenticator> mfas = new HashMap<>();
		mfas.put(GoogleAuthenticator.TYPE, new GoogleAuthenticator(new SoftwareServerCryptoModule(new Random(1))));
		dummyIdP = new TestPasswordJWTIdP(new TestInMemoryStorage() ,mfas);
	}
	

	@Test
	public void testCreateUser() throws UserCreationFailedException {
		List<PasswordJWTIdP> servers = new LinkedList<>();
		servers.add(dummyIdP);
		PasswordJWTClient client = new PasswordJWTClient(servers);
		client.createUser(user, password);
	}

	
	@Test
	public void testCreateUserAndProveIdentity() throws UserCreationFailedException {
		List<PasswordJWTIdP> servers = new LinkedList<>();
		servers.add(dummyIdP);
		PasswordJWTClient client = new PasswordJWTClient(servers);
		client.createUserAndProveIdentity(user, password, new TestIdentityProof(signature, attributes));
	}
	
	@Test
	public void testProveIdentity() throws UserCreationFailedException {
		List<PasswordJWTIdP> servers = new LinkedList<>();
		servers.add(dummyIdP);
		PasswordJWTClient client = new PasswordJWTClient(servers);
		client.proveIdentity(user, password, new TestIdentityProof(signature, attributes), null, "NONE");
	}
	
	@Test
	public void testAuthenticate() throws UserCreationFailedException, AuthenticationFailedException {
		List<PasswordJWTIdP> servers = new LinkedList<>();
		servers.add(dummyIdP);
		PasswordJWTClient client = new PasswordJWTClient(servers);
		Policy policy = new Policy();
		List<Predicate> predicates = new ArrayList<>();
		for(String s: claims) {
			Predicate predicate = new Predicate();
			predicate.setAttributeName(s);
			predicate.setOperation(Operation.REVEAL);
			predicates.add(predicate);
		}
		policy.setPredicates(predicates);
		String reply = client.authenticate(user, password, policy, null, "NONE");
		assertThat(reply, is("Dummy-Authenticate"));
	}
	
	@Test
	public void testRequestMFAChallenge() throws Exception {
		List<PasswordJWTIdP> servers = new LinkedList<>();
		servers.add(dummyIdP);
		PasswordJWTClient client = new PasswordJWTClient(servers);
		String challenge = client.requestMFAChallenge(user, password, GoogleAuthenticator.TYPE);
		assertEquals(challenge, this.challenge);
	}
	
	@Test(expected = AuthenticationFailedException.class)
	public void testMissingConfirmMFA() throws Exception {
		List<PasswordJWTIdP> servers = new LinkedList<>();
		servers.add(dummyIdP);
		PasswordJWTClient client = new PasswordJWTClient(servers);

		client.confirmMFA(user, password, "none", GoogleAuthenticator.TYPE);
		fail();
	}

	@Test(expected = AuthenticationFailedException.class)
	public void testMissingRemoveMFA() throws Exception {
		List<PasswordJWTIdP> servers = new LinkedList<>();
		servers.add(dummyIdP);
		PasswordJWTClient client = new PasswordJWTClient(servers);

		client.removeMFA(user, password, "none", GoogleAuthenticator.TYPE);
		fail();
	}

	@Test
	public void testMissingUser() {
		List<PasswordJWTIdP> servers = new LinkedList<>();
		servers.add(dummyIdP);
		PasswordJWTClient client = new PasswordJWTClient(servers);
		try {
			client.removeMFA("otherUser", password, "none", GoogleAuthenticator.TYPE);
			fail();
		} catch (Exception e) {
			// correct behaviour
		}
	}

	@Test(expected = AuthenticationFailedException.class)
	public void testFailedAuthentication() throws Exception {
		List<PasswordJWTIdP> servers = new LinkedList<>();
		servers.add(dummyIdP);
		PasswordJWTClient client = new PasswordJWTClient(servers);

		client.authenticate(user, "wrong-password", new Policy(), null, "NONE");
		fail();
	}

	@Test(expected = AuthenticationFailedException.class)
	public void testFailedDeleteAttributes() throws Exception {
		List<PasswordJWTIdP> servers = new LinkedList<>();
		servers.add(dummyIdP);
		PasswordJWTClient client = new PasswordJWTClient(servers);
		client.deleteAttributes(user, "wrong-password", Arrays.asList("name"), null, "NONE");
		fail();
	}

	@Test(expected = AuthenticationFailedException.class)
	public void testFailedDeleteAccount() throws Exception {
		List<PasswordJWTIdP> servers = new LinkedList<>();
		servers.add(dummyIdP);
		PasswordJWTClient client = new PasswordJWTClient(servers);
		client.deleteAccount(user, "wrong-password", null, "NONE");
		fail();
	}
	
	@Test
	public void testGetAllAttributes() throws Exception {
		List<PasswordJWTIdP> idps = new ArrayList<PasswordJWTIdP>();
		TestIdP idp = new TestIdP();
		idps.add(idp);
		PasswordJWTClient authClient = new PasswordJWTClient(idps) {
			@Override
			public String authenticate(String username, String password, Policy policy, String token, String type) {
				//Not used for testing
				return null;
			}
		};
		idp.createUser(new PasswordAuthentication("username", "password"));
		Map<String, Attribute> attributes = authClient.getAllAttributes("username", "password", null, "NONE");
		assertEquals(2, attributes.size());
		assertEquals(new Attribute("John"), attributes.get("name"));
		assertEquals(new Attribute("John2"), attributes.get("name2"));
	}

	@Test
	public void testGetAllAttributesWithMFA() throws Exception {
		List<PasswordJWTIdP> idps = new ArrayList<PasswordJWTIdP>();
		TestIdP idp = new TestIdP();
		idps.add(idp);
		PasswordJWTClient authClient = new PasswordJWTClient(idps) {
			@Override
			public String authenticate(String username, String password, Policy policy, String token, String type) {
				//Not used for testing
				return null;
			}
		};
		idp.createUser(new PasswordAuthentication("username", "password"));
		Map<String, Attribute> attributes = authClient.getAllAttributes("username", "password", "TOKEN", GoogleAuthenticator.TYPE);
		assertEquals(2, attributes.size());
		assertEquals(new Attribute("John"), attributes.get("name"));
		assertEquals(new Attribute("John2"), attributes.get("name2"));
	}
	
	@Test(expected = AuthenticationFailedException.class)
	public void testGetAllAttributesServerFailure() throws Exception {
		List<PasswordJWTIdP> idps = new ArrayList<PasswordJWTIdP>();
		TestIdP idp = new TestIdP() {
			@Override
			public Map<String, Attribute> getAllAttributes(PasswordAuthentication authenticationData, byte[] cookie) {
				throw new RuntimeException("simulated server failure");
			}
		};
		idps.add(idp);
		PasswordJWTClient authClient = new PasswordJWTClient(idps) {
			@Override
			public String authenticate(String username, String password, Policy policy, String token, String type) {
				//Not used for testing
				return null;
			}
		};
		authClient.getAllAttributes("username", "password", null, "NONE");
		fail();
	}

	@Test
	public void testDeleteAttributes() throws Exception {
		List<PasswordJWTIdP> idps = new ArrayList<PasswordJWTIdP>();
		TestIdP idp = new TestIdP();
		idps.add(idp);
		PasswordJWTClient authClient = new PasswordJWTClient(idps) {
			@Override
			public String authenticate(String username, String password, Policy policy, String token, String type) {
				//Not used for testing
				return null;
			}
		};
		idp.createUser(new PasswordAuthentication("username", "password"));
		List<String> toDelete = new ArrayList<String>();
		toDelete.add("John");
		authClient.deleteAttributes("username", "password", toDelete, null, "NONE");
		assertTrue(idp.deleteAttributeReached);
	}

	@Test
	public void testDeleteAttributesWithMFA() throws Exception {
		List<PasswordJWTIdP> idps = new ArrayList<PasswordJWTIdP>();
		TestIdP idp = new TestIdP();
		idps.add(idp);
		PasswordJWTClient authClient = new PasswordJWTClient(idps) {
			@Override
			public String authenticate(String username, String password, Policy policy, String token, String type) {
				//Not used for testing
				return null;
			}
		};
		idp.createUser(new PasswordAuthentication("username", "password"));
		List<String> toDelete = new ArrayList<String>();
		toDelete.add("John");
		authClient.deleteAttributes("username", "password", toDelete, "TOKEN", GoogleAuthenticator.TYPE);
		assertTrue(idp.deleteAttributeReached);
	}
	
	@Test(expected = AuthenticationFailedException.class)
	public void testDeleteAttributesServerFailure() throws Exception {
		List<PasswordJWTIdP> idps = new ArrayList<PasswordJWTIdP>();
		TestIdP idp = new TestIdP() {
			@Override
			public boolean deleteAttribute(PasswordAuthentication authenticationData, byte[] cookie, List<String> attributes) {
				throw new RuntimeException("simulated server failure");
			}
		};
		idps.add(idp);
		PasswordJWTClient authClient = new PasswordJWTClient(idps) {
			@Override
			public String authenticate(String username, String password, Policy policy, String token, String type) {
				//Not used for testing
				return null;
			}
		};
		List<String> toDelete = new ArrayList<String>();
		toDelete.add("John");
		authClient.deleteAttributes("username", "password", toDelete, null, "NONE");
		fail();
	}
	
	@Test(expected = AuthenticationFailedException.class)
	public void testDeleteAttributesServerNegativeAnswer() throws Exception {
		List<PasswordJWTIdP> idps = new ArrayList<PasswordJWTIdP>();
		TestIdP idp = new TestIdP() {
			@Override
			public boolean deleteAttribute(PasswordAuthentication authenticationData, byte[] cookie, List<String> attributes) {
				return false;
			}
		};
		idps.add(idp);
		PasswordJWTClient authClient = new PasswordJWTClient(idps) {
			@Override
			public String authenticate(String username, String password, Policy policy, String token, String type) {
				//Not used for testing
				return null;
			}
		};

		List<String> toDelete = new ArrayList<String>();
		toDelete.add("John");
		authClient.deleteAttributes("username", "password", toDelete, null, "NONE");
		fail();
	}
	
	@Test
	public void testDeleteAccountWithMFA() throws Exception {
		List<PasswordJWTIdP> idps = new ArrayList<PasswordJWTIdP>();
		TestIdP idp = new TestIdP();
		idps.add(idp);
		PasswordJWTClient authClient = new PasswordJWTClient(idps) {
			@Override
			public String authenticate(String username, String password, Policy policy, String token, String type) {
				//Not used for testing
				return null;
			}
		};

		idp.createUser(new PasswordAuthentication("username", "password"));
		authClient.deleteAccount("username", "password", "TOKEN", GoogleAuthenticator.TYPE);
		assertTrue(idp.deleteAccountReached);
	}

	@Test(expected = AuthenticationFailedException.class)
	public void testDeleteAccountServerFailure() throws Exception {
		List<PasswordJWTIdP> idps = new ArrayList<PasswordJWTIdP>();
		TestIdP idp = new TestIdP() {
			@Override
			public boolean deleteAccount(PasswordAuthentication authenticationData, byte[] cookie) {
				throw new RuntimeException("simulated server failure");
			}
		};
		idps.add(idp);
		PasswordJWTClient authClient = new PasswordJWTClient(idps) {
			@Override
			public String authenticate(String username, String password, Policy policy, String token, String type) {
				//Not used for testing
				return null;
			}
		};

		authClient.deleteAccount("username", "password", null, "NONE");
		fail();
	}
	
	@Test(expected = AuthenticationFailedException.class)
	public void testDeleteAccountServerFailureNice() throws Exception {
		List<PasswordJWTIdP> idps = new ArrayList<PasswordJWTIdP>();
		TestIdP idp = new TestIdP() {
			@Override
			public boolean deleteAccount(PasswordAuthentication authenticationData, byte[] cookie) {
				return false;
			}
		};
		idps.add(idp);
		PasswordJWTClient authClient = new PasswordJWTClient(idps) {
			@Override
			public String authenticate(String username, String password, Policy policy, String token, String type) {
				//Not used for testing
				return null;
			}
		};

		authClient.deleteAccount("username", "password", null, "NONE");
		fail();
	}
	
	@Test
	public void testChangePasswordWithMFA() throws Exception {
		List<PasswordJWTIdP> idps = new ArrayList<PasswordJWTIdP>();
		TestIdP idp = new TestIdP();
		idps.add(idp);
		PasswordJWTClient authClient = new PasswordJWTClient(idps) {
			@Override
			public String authenticate(String username, String password, Policy policy, String token, String type) {
				//Not used for testing
				return null;
			}
		};
		idp.createUser(new PasswordAuthentication("username", "password"));
		authClient.changePassword("username", "password", "password2", "TOKEN", GoogleAuthenticator.TYPE);
		assertTrue(idp.changePWReached);
	}
	
	@Test(expected=AuthenticationFailedException.class)
	public void testChangePasswordBadServerSignature() throws Exception {
		List<PasswordJWTIdP> idps = new ArrayList<PasswordJWTIdP>();
		TestIdP idp = new TestIdP() {
			@Override
			public void changePassword(PasswordAuthentication oldAuthenticationData, PasswordAuthentication newAuthenticationData, byte[] cookie) {
				throw new RuntimeException("planned failure");
			}
		};
		idps.add(idp);
		PasswordJWTClient authClient = new PasswordJWTClient(idps) {
			@Override
			public String authenticate(String username, String password, Policy policy, String token, String type) {
				//Not used for testing
				return null;
			}
		};
		authClient.changePassword("username", "password", "password2", null, "NONE");
		fail();
	}
	
	private class TestIdP extends PasswordJWTIdP {

		public TestIdP() throws Exception {
			super(new InMemoryUserPasswordDatabase(), new ArrayList<>(), new HashMap<String, MFAAuthenticator>());
		}
		
		@Override
		public String generateSession(PasswordAuthentication authentication, String token, String type)
				throws AuthenticationFailedException {
			if("NONE".equals(type)) {
				return "Y29va2ll";
			}
			assertEquals("TOKEN", token);
			assertEquals(GoogleAuthenticator.TYPE, type);
			return "Y29va2ll";
		}
		
		@Override
		public Certificate getCertificate() {
			return TestParameters.getRSA1Cert();
		}
	
		@Override
		public Map<String, Attribute> getAllAttributes(PasswordAuthentication authenticationData, byte[] cookie) {
			assertEquals("username", authenticationData.getUsername());
			assertEquals("cookie", new String(cookie));
			
			Map<String, Attribute> attributes = new HashMap<String, Attribute>();
			attributes.put("name", new Attribute("John"));
			attributes.put("name2", new Attribute("John2"));
			return attributes;
		}
		
		public boolean deleteAttributeReached = false;
		@Override
		public boolean deleteAttribute(PasswordAuthentication authenticationData, byte[] cookie, List<String> attributes) {
			assertEquals("username", authenticationData.getUsername());
			assertEquals("cookie", new String(cookie));
			assertEquals(1, attributes.size());
			assertEquals("John", attributes.get(0));
			deleteAttributeReached = true;
			return true;
		}
		
		public boolean deleteAccountReached = false;
		@Override
		public boolean deleteAccount(PasswordAuthentication authenticationData, byte[] cookie) {
			assertEquals("cookie", new String(cookie));
			assertEquals("username", authenticationData.getUsername());
			deleteAccountReached = true;
			return true;
		}
		
		public boolean changePWReached = false;
		@Override
		public void changePassword(PasswordAuthentication oldAuthenticationData, PasswordAuthentication newAuthenticationData, byte[] cookie) {
			assertEquals("cookie", new String(cookie));
			assertEquals("username", oldAuthenticationData.getUsername());
			assertEquals("password2", newAuthenticationData.getPassword());
			changePWReached = true;
		}
	}
}
